/*
  m1rand version 2.0
  item randomizer

  Recent changes:
    Added more_upgrades option.
    Should now be able to generate more different patterns.

  To run m1.nes is required to be in same directory.
  Then m1r.nes is created and will be overwritten if it exists.

  Command line options
  p, pretend - Just pretend, don't write anything to disk.
  s, save_seed - Saves the seed number to the file name.
  u, more_upgrades - Replacing some missiles with random upgrades.
  X - X is the seed number you want to use (not an actual X).


  To compile on linux run:
  gcc -o m1rand.bin -Wall -pedantic m1rand.c

*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>


/* swap unsigned chars*/
void swap_uc(unsigned char *val1, unsigned char *val2){
  unsigned char tmp=0;
  tmp = *val1;
  *val1 = *val2;
  *val2 = tmp;
  return;
}

/* swap ints */
void swap_int(int *val1, int *val2){
  unsigned char tmp=0;
  tmp = *val1;
  *val1 = *val2;
  *val2 = tmp;
  return;
}

/* check if target location doesn't require key item */
int chk_req( int *array_size, unsigned char req[], unsigned char *key_item ){
  int i = 0;
  for ( i=0; i < *array_size; i++){
    if ( req[i] == *key_item ){
      return 1;
    }
  }
  return 0;
}

/* check if item is collectable */
/*args: col_value, item_required[ key_addr[k] ], item_required2[ key_addr[k] ], col_array_size, item_req_array_size*/
int item_col ( unsigned char col_value[], unsigned char req[], unsigned char req2[], int col_array_size, int item_req_array_size){
  int i = 0;
  int j = 0;
  int req_met = 0;
  int req2_used = 0;
  req_met = 0;
  
  for ( i = 0; i < item_req_array_size; i++ ){
    for ( j = 0; j < col_array_size; j++ ){
      if ( req[i] == col_value[j] || req[i] == 0xFF ){
	/* requirement i met. all requirements must be */
	/* printf("req %02X : ", req[i]); */
	req_met++;
	break;
      }
    }
  }
  /* printf("\n req_met %d\n", req_met); */
  if ( req_met == item_req_array_size ){
    /* printf("req1 is good \n"); */
    return 1;
  }
  
  /* check if any req2 is used */
  for ( i = 0; i < item_req_array_size; i++ ){
    if ( req2[i] != 0xFF ){
      req2_used = 1;
    }
  }

  /* check if req2 requirements are met. */
  if ( req2_used == 1 ){
    req_met = 0;
    for ( i = 0; i < item_req_array_size; i++ ){
      for ( j = 0; j < col_array_size; j++ ){
	if ( req2[i] == col_value[j] || req2[i] == 0xFF ){
	  /* requirement i met. all requirements must be */
	  req_met++;
	  break;
	}
      }
    }
    if ( req_met == item_req_array_size ){
      /* printf("req2 is good \n"); */
      return 1;
    }
  }

  return 0;
}


/* key_value[k], col_value, col_array_size */
/* check if item is allready collected */
int item_collected( unsigned char *key_value, unsigned char *col_value, int col_array_size){
  int i = 0;
  for ( i = 0; i < col_array_size; i++ ){
    if ( *key_value == col_value[i] ){
      return 1;
    }
  }
  return 0;
}


int main(int argc, char *argv[] ){
  /* const int times_to_randomize=5; */
  unsigned int seed=0;
  char *arg_str;
  char file_to_save[80];

  /* options */
  int save_seed_num = 0;
  int pretend = 0;
  int more_upgrades = 0;

  int replacable_missiles = 11;


  int randomize_extra=5;
  int filesize=0;
  int errnum;
  FILE *ptr_origfile;
  unsigned char *buffer;

  FILE *ptr_newfile;


  /* moving items init */

  int i=0,j=0,k=0;
  int rand_num=0;
  /* int continue_loop=0; */



  /* check if key items are collectable */
  int col_num = 0;/* number of items collected */
  int col_array_size = 4;
  unsigned char col_value[ col_array_size ];
  col_value[0] = 0xFF;
  col_value[1] = 0xFF;
  col_value[2] = 0xFF;
  col_value[3] = 0xFF;
  
  int item_array_size=37;
  int item_addr[item_array_size];
  unsigned char item_value[item_array_size];

  int item_req_array_size=4;
  unsigned char item_required[item_array_size][item_req_array_size];
  unsigned char item_required2[item_array_size][item_req_array_size];


  for (j=0;j<item_array_size;j++){
    for (k=0;k<item_req_array_size;k++){
      item_required[j][k]=0xFF;
    }
  }

  for (j=0;j<item_array_size;j++){
    for (k=0;k<item_req_array_size;k++){
      item_required2[j][k]=0xFF;
    }
  }

  
  /* calling all items that can be required "key" */
  int key_array_size=4;
  int key_addr[key_array_size];/* number is array number */
  unsigned char key_value[key_array_size];


  /* Brinstar */
  /* first etank (requires freeze or highjump n bombs) */
  item_addr[0]=0x6446;
  item_value[0]=0x08;
  item_required[0][2]=0x07;

  /* first missile in golden brinstar near norfair entrance */
  item_addr[1]=0x6432;
  item_value[1]=0x09;

  /* actual first etank in brinstar on the way towards bombs. */
  item_addr[2]=0x641C;
  item_value[2]=0x08;

  /* bombs req missile */
  item_addr[3]=0x640F;
  item_value[3]=0x00;
  item_required[3][3]=0x09;
  
  /* first ice beam in brinstar(require missile and bombs) */
  item_addr[4]=0x6425;
  item_value[4]=0x07;
  item_required[4][0]=0x00;
  item_required[4][3]=0x09;
  
  /* varia suit. (require high jump, bombs and missile) */
  item_addr[5]=0x63F1;
  item_value[5]=0x05;
  item_required[5][0]=0x00;
  item_required[5][1]=0x01;
  item_required[5][3]=0x09;
  
  /* missile after varia suit(require bombs) */
  item_addr[6]=0x63FA;
  item_value[6]=0x09;
  item_required[6][0]=0x00;
  
  /* e tank near varia suit(require bombs) */
  item_addr[7]=0x6400;
  item_value[7]=0x08;
  item_required[7][0]=0x00;
  
  /* long beam(req missile) */
  item_addr[8]=0x6409;
  item_value[8]=0x02;
  item_required[8][3]=0x09;

  
  /* Norfair 12m 1e */
  /* Missile first to the left when entering from brinstar. */
  item_addr[9]=0xA329;
  item_value[9]=0x09;
  
  /* Missile second to the left when entering from brinstar.(require bombs) */
  item_addr[10]=0xA332;
  item_value[10]=0x09;
  item_required[10][0]=0x00;
  
  /* Missile first to the right down left when entering from brinstar.(require bombs and ice beam) */
  item_addr[11]=0xA33F;
  item_value[11]=0x09;
  item_required[11][0]=0x00;
  item_required[11][2]=0x07;

  /* Missile second to the right down left when entering from brinstar.(require bombs and ice beam) */
  item_addr[12]=0xA339;
  item_value[12]=0x09;
  item_required[12][0]=0x00;
  item_required[12][2]=0x07;

  /* ice beam norfair (req icebeam, bombs and missile) or (bombs and high jump) */
  item_addr[13]=0xA318;
  item_value[13]=0x07;
  item_required[13][0]=0x00;
  item_required[13][2]=0x07;
  item_required[13][3]=0x09;
  item_required2[13][0]=0x00;
  item_required2[13][1]=0x01;

  /* Missile first 1 level above ice beam norfair (req icebeam or high jump)  */
  item_addr[14]=0xA30F;
  item_value[14]=0x09;
  item_required[14][2]=0x07;
  item_required2[14][1]=0x01;

  /* Missile second 1 level above ice beam (req icebeam or high jump) */
  item_addr[15]=0xA309;
  item_value[15]=0x09;
  item_required[15][2]=0x07;
  item_required2[15][1]=0x01;

  /* Missile third 1 level above ice beam (req icebeam or high jump) */
  item_addr[16]=0xA303;
  item_value[16]=0x09;
  item_required[16][2]=0x07;
  item_required2[16][1]=0x01;

  /* Missile first 2 levels above ice beam (req icebeam, bombs and missile) or (bombs and high jump) */
  item_addr[17]=0xA2F5;
  item_value[17]=0x09;
  item_required[17][0]=0x00;
  item_required[17][2]=0x07;
  item_required[17][3]=0x09;
  item_required2[17][0]=0x00;
  item_required2[17][1]=0x01;


  /* Missile second 2 levels above ice beam (req icebeam, bombs and missile) or (bombs and high jump) */
  item_addr[18]=0xA2EF;
  item_value[18]=0x09;
  item_required[18][0]=0x00;
  item_required[18][2]=0x07;
  item_required[18][3]=0x09;
  item_required2[18][0]=0x00;
  item_required2[18][1]=0x01;



  /* high jump (req missile and bombs) */
  item_addr[19]=0xA370;
  item_value[19]=0x01;
  item_required[19][0]=0x00;
  item_required[19][3]=0x09;

  /* screw attack (req missile, bombs and ice beam) */
  item_addr[20]=0xA351;
  item_value[20]=0x03;
  item_required[20][0]=0x00;
  item_required[20][2]=0x07;
  item_required[20][3]=0x09;

  /* etank down right of screw attack (require bombs, missile and ice beam) */
  item_addr[21]=0xA383;
  item_value[21]=0x08;
  item_required[21][0]=0x00;
  item_required[21][2]=0x07;
  item_required[21][3]=0x09;

  /* Wave beam (req bombs, missile and ice beam) */
  item_addr[22]=0xA39F;
  item_value[22]=0x06;
  item_required[22][0]=0x00;
  item_required[22][2]=0x07;
  item_required[22][3]=0x09;

  /* Missile right of wave beam (req bombs, missile and ice beam) */
  item_addr[23]=0xA396;
  item_value[23]=0x09;
  item_required[23][0]=0x00;
  item_required[23][2]=0x07;
  item_required[23][3]=0x09;

  /* Missile first one under wave beam (req bombs, missile and ice beam) */
  item_addr[24]=0xA3AD;
  item_value[24]=0x09;
  item_required[24][0]=0x00;
  item_required[24][2]=0x07;
  item_required[24][3]=0x09;

  /* Missile second one under wave beam (req bombs, missile and ice beam) */
  item_addr[25]=0xA3B3;
  item_value[25]=0x09;
  item_required[25][0]=0x00;
  item_required[25][2]=0x07;
  item_required[25][3]=0x09;

  
  /* Lower Norfair */
  /* Missile first to left up left when entering from norfair (req bombs) */
  item_addr[26]=0x16223;
  item_value[26]=0x09;
  item_required[26][0]=0x00;

  /* Etank left when entering from norfair (req bombs and missile) */
  item_addr[27]=0x16231;
  item_value[27]=0x08;
  item_required[27][0]=0x00;
  item_required[27][3]=0x09;

  /* Missile to the right of hard jump usually with unmorph jump. (req bombs and missile) */
  item_addr[28]=0x1623A;
  item_value[28]=0x09;
  item_required[28][0]=0x00;
  item_required[28][3]=0x09;

  /* Etank behind Riddley. (req bombs and missile) */
  item_addr[29]=0x16243;
  item_value[29]=0x08;
  item_required[29][0]=0x00;
  item_required[29][3]=0x09;

  /* Missile in bottom level. (req bombs) */
  item_addr[30]=0x1624C;
  item_value[30]=0x09;
  item_required[30][0]=0x00;
  
  /* Kraids Lair */
  /* Missile first to the left when entering from brinstar (req bombs) */
  item_addr[31]=0x12293;
  item_value[31]=0x09;
  item_required[31][0]=0x00;

  /* Missile one level above fake Kraid. (req bombs) */
  item_addr[32]=0x122B4;
  item_value[32]=0x09;
  item_required[32][0]=0x00;
  
  /* Missile first to the right when entering from brinstar. (req bombs and missile) */
  item_addr[33]=0x12299;
  item_value[33]=0x09;
  item_required[33][0]=0x00;
  item_required[33][3]=0x09;

  /* Etank first to right when entering from brinstar. (req bombs and missile)  */
  item_addr[34]=0x122A2;
  item_value[34]=0x08;
  item_required[34][0]=0x00;
  item_required[34][3]=0x09;

  /* Missile 3 levels down right down left */
  item_addr[35]=0x122AB;
  item_value[35]=0x09;
  item_required[35][0]=0x00;
  item_required[35][3]=0x09;

  /* Etank at Kraid. */
  item_addr[36]=0x122C4;
  item_value[36]=0x08;
  item_required[36][0]=0x00;
  item_required[36][3]=0x09;



  
  
/* KEY items */
  key_value[0] = 0x00;/* bombs */
  key_value[1] = 0x01;/* high jump */
  key_value[2] = 0x07;/* ice beam */
  key_value[3] = 0x09;/* missile */

  /* Upgrades */
  unsigned char upgrade_value[7];
  int upgrade_array_size = 7;
  
  upgrade_value[0] = 0x00;/* bombs */
  upgrade_value[1] = 0x01;/* high jump */
  upgrade_value[2] = 0x02;/* long beam */
  upgrade_value[3] = 0x03;/* screw attack */
  upgrade_value[4] = 0x05;/* varia suit */
  upgrade_value[5] = 0x06;/* wave beam */
  upgrade_value[6] = 0x07;/* ice beam */

  
  printf("m1rand version 2.0\n\n");
  printf("Trying to read m1.nes\n");
  
  ptr_origfile = fopen("m1.nes","rb");
  if (ptr_origfile == NULL){
    errnum = errno;
    fprintf(stderr, "Value of errno: %d\n", errno);
    perror("Error printed by perror");
    fprintf(stderr, "Error opening file: %s\n", strerror(errnum));
    printf("\n m1.nes is required to be in same directory.\n Then m1r.nes is created. Will be overwritten if exists.");
    return 1;
  }
  else {
    fseek(ptr_origfile, 0, SEEK_END);

    filesize = ftell(ptr_origfile);
    printf("Filesize: %d bytes.\n", filesize);
    if (filesize == 131088){
      printf("Found the original m1.nes\n");
    }
    else{
      printf("m1.nes is required to be in same directory.\n Then m1r.nes is created and will be overwritten if it exists.");
    }

    rewind(ptr_origfile);

    buffer = (unsigned char *) malloc(sizeof(unsigned char) * filesize+1);

    if(fread(buffer, sizeof(unsigned char), filesize, ptr_origfile) != filesize){
      printf("error reading file to buffer.\n");
    }
    buffer[filesize] = '\0';
    
    fclose (ptr_origfile);

    
    /* reading user arguments and userdefined seed */
    sprintf( file_to_save, "m1r.nes");
    printf("\nnumber of args %d\n", argc);
    if(argc > 1){
      for(i=1; i < argc; i++){
	if ( strcmp(argv[i], "save_seed") == 0 || strcmp(argv[i], "s") == 0 ){
	  printf("Writing to file m1s[seednum].nes instead of m1r.nes\n");
	  save_seed_num = 1;
	}
	else if ( strcmp(argv[i], "pretend") == 0 || strcmp(argv[i], "p") == 0 ){
	  printf("Only pretending. Not writing any file.\n");
	  pretend = 1;
	}
	else if ( strcmp(argv[i], "more_upgrades") == 0 || strcmp(argv[i], "u") == 0 ){
	  printf("Replacing some missiles with random upgrades.\n");
	  more_upgrades = 1;
	}
	else{
	  /* arg_str = (char *) malloc(sizeof(char) * 150); */
	  seed = strtoul(argv[i], &arg_str, 10);
	  printf("Using userdefined seed: %u\n", seed);
	}
      }
    }

    
    if ( seed == 0 ){
      printf("Generating a seed based on time (seconds since Epoch).\n");
      seed = time(NULL);
    }

    if ( save_seed_num == 1 ){
      sprintf( file_to_save, "m1r%d.nes", seed);
      printf("Your file will be saved as %s \n", file_to_save);
    }

    

    printf("Your seed number is: %d\n", seed);
    /* Generating random numbers based on the seed. Should only be run once. */
    srand(seed);

    
    if ( more_upgrades == 1 ){
      for (i=0; i < item_array_size; i++){
	if ( item_value[i] == 0x09 ){
	  rand_num = rand() % (upgrade_array_size );
	  item_value[i] = upgrade_value[ rand_num ];
	  replacable_missiles--;
	}
	if ( replacable_missiles < 1 ){
	  break;
	}
      }
    }
    

    /* swap items arround a bit extra. */
    /* items seems to have a high chance of using their original possition without it. */
    for (j = 0; j < randomize_extra; j++){
      for(i=0; i<item_array_size; i++){
	rand_num = rand() % (item_array_size );
	swap_uc(&item_value[i], &item_value[rand_num]);
      }
    }

    while ( 1 ){
    
      /* swap items arround */
      for(i=0; i<item_array_size; i++){
	rand_num = rand() % (item_array_size );
	
	swap_uc(&item_value[i], &item_value[rand_num]);
      }

      for (i=0; i < key_array_size; i++){
	key_addr[i] = -1;
      }
      
      /* Find key items */
      for(i=0; i<item_array_size; i++){
	for(j=0; j<key_array_size; j++){
	  rand_num = rand() % (item_array_size );
	  if(item_value[rand_num] == key_value[j]){
            /* adding randomness so that none uniq key items doesn't allways gets the last one picked. */
	    key_addr[j] = rand_num;
	  }
	  if(item_value[i] == key_value[j] && key_addr[j] == -1 ){
	    key_addr[j] = i;
	  }

	}
      }
      
      /* check if key items are collectable */
      /* i is item collection run */

      col_value[0] = 0xFF;
      col_value[1] = 0xFF;
      col_value[2] = 0xFF;
      col_value[3] = 0xFF;
      col_num = 0;

      for ( i = 0; i < item_array_size; i++ ){
	for ( k = 0; k < key_array_size; k++ ){
	  if ( item_collected( &key_value[k], col_value, col_array_size) == 1 ) {
	    continue;
	  }
	  /* if ( (item_collected( &key_value[3] , col_value, col_array_size) == 1 ) && key_addr[0] == 8){/\* missiles never get collected if key_addr[0] == 8 ???? *\/ */
	  /*   printf ("item_col( col_value, item_required[ key_addr[ k %d ] ], item_required2[ key_addr[ k %d ] ], col_array_size %d, item_req_array_size %d\n", k, k, col_array_size, item_req_array_size); */
	  /* } */
	  if ( item_col( col_value, item_required[ key_addr[k] ], item_required2[ key_addr[k] ], col_array_size, item_req_array_size ) == 1 ){
	    col_value[ col_num ] = key_value[k];
	    /* printf ("success item_col( col_value, item_required[ key_addr[ k %d ] ], item_required2[ key_addr[ k %d ] ], col_array_size %d, item_req_array_size %d\n", k, k, col_array_size, item_req_array_size); */

	    /* printf( "Item %02X collected\n", key_value[k]); */
	    col_num++;continue;
	  }
	}
      }
      if ( col_num != key_array_size ){
	/* printf("Only collected %d, and the seed is %d\n", col_num, seed); */
	/* break; */
      }
      else{
	break;/* the items can be collected so no need to move them again */
      }
    }
    
    
    /* for (i=0; i< key_array_size;i++){ */
    /*   printf("key:%d addr:%d \n", i, key_addr[i]); */
    /* } */
    
    
    /* Save items to buffer for writing. */
    for(i=0; i<item_array_size; i++){
      buffer[item_addr[i]] = item_value[i];
    }



    /* writing the new file */
    if ( pretend == 0 ){
      ptr_newfile = fopen( file_to_save, "wb");
      if (ptr_newfile == NULL){
	errnum = errno;
	fprintf(stderr, "Value of errno: %d\n", errno);
	perror("Error printed by perror");
	fprintf(stderr, "Error opening file: %s\n", strerror(errnum));
	return 1;
      }
      else {
	fwrite(buffer, sizeof(unsigned char), filesize, ptr_newfile);
	fclose (ptr_newfile);
	printf( "%s written.\n", file_to_save);
      }
    }
  }
  free(buffer);
  /* printf("I remember!\n"); */
  return 0;
}
